<?php

/**
* SmsSender class
* This class is able to send short messages using the Gigahost Short Message Gateway API.
*
* @version 0.6
* @author Gigahost <support@gigahost.dk>
*/
class SmsSender {
    
    // The following constants shouldn't be modified unless you know what you are doing.
    const SERVICE_ADDRESS = '_sms._tcp.tel.gigahost.dk';
    const MAX_RECIPIENTS = 1000;
    
    private static $validStatusCodes = array(200, 201, 202);
    
    private $username, $password;
    
    public function __construct($username,$password) {
        $this->username = $username;
        $this->password = $password;
    }
    
    /** 
    * Sends a message using the Short Message gateway.
    * @param string $sender The number that will be the sender of the SMS, including the country code. Example: +4542031337
    * @param mixed $recipients Either one recipient specified as a string, or an array of recipients as strings. The number should be formatted with the country code. Example: +4542031337
    * @param string The message text.
    * @return Associative array with information about the message.
    */
    public function sendMessage($sender,$recipients,$message,$debug=false) {
        if (!is_array($recipients)) $recipients = array($recipients);
        if (count($recipients)>self::MAX_RECIPIENTS) throw new Exception("Unable to send messages to more than " . self::MAX_RECIPIENTS . " recipients");
        $recipientsString = implode(',',$recipients);

        $res = $this->doRequest('messages','POST',array('sender'=>$sender,'recipients'=>$recipientsString,'message'=>$message),$debug);
        $xml = simplexml_load_string(trim($res));

        $erroneousRecipients = array();
        if ($xml->erroneousRecipients->erroneousRecipient) {
            foreach($xml->erroneousRecipients->erroneousRecipient as $erroneousRecipient) {
                array_push($erroneousRecipients,(string)$erroneousRecipient);
            }
        }
        
        $message = array();
        $message['id']=(string)$xml->Message['id'];
        $message['cost']=(int)$xml->cost;
        $message['acceptedRecipientCount']=(int)$xml->acceptedRecipientCount;
        $message['erroneousRecipientCount']=(int)$xml->erroneousRecipientCount;
        $message['erroneousRecipients'] = $erroneousRecipients;
        
        return $message;

    }

	/**
	 * Looks up a message according to uuid
	 * @param string $messageUuid The unique message uuid 
	 * @return Associative array with information about the message.
	 */ 
	public function getMessageStatus($messageUuid){
		try {
			$res = $this->doRequest('messages/'.$messageUuid,'GET');
	        $xml = simplexml_load_string(trim($res));			
			$message = array(
				'id' => (String)$xml->id, 
				'senderText'=>(String)$xml->senderText, 
				'message'=>(String)$xml->message, 
				'timeEnqueued'=>strtotime((String)$xml->timeEnqueued),
				'messageRecipients' => array()
			);
			foreach($xml->MessageRecipients->MessageRecipient as $mr) 
			array_push($message['messageRecipients'], 
							array('number'=>(String)$mr['number'], 
							'erroneous'=>((String)$mr['erroneous']=='yes'?true:false),
							'timeSent'=>strtotime((String)$mr['timeSent']),
							'timeDelivered'=>strtotime((String)$mr['timeDelivered'])));
			return $message;
		} catch (Exception $e) {
			if(strpos($e->getMessage(), "404")) throw new Exception('Message not found.');
			
		}
	}
    
    private function doRequest($request, $type, array $parameters = null,$debug=false){
        $host = self::getServiceAddress();
        if (!$host) throw new Exception("No message servers are currently available, and messages cannot be sent.");
        
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_USERPWD, $this->username.':'.$this->password);
        curl_setopt($ch, CURLOPT_HEADER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_HTTPHEADER, array('Accept: application/xml'));
        
        $url = 'https://'.$host.'/'.$request;
        if(count($parameters)) $query = http_build_query($parameters); else $query = '';
        
        switch($type){
            case 'GET': 
                curl_setopt($ch, CURLOPT_URL, $url.($query?'?'.$query:''));
            break;
            case 'POST':
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_POST, count($parameters));
                curl_setopt($ch, CURLOPT_POSTFIELDS, $query);
            break;
            case 'PUT':
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'PUT');  
                curl_setopt($ch, CURLOPT_POSTFIELDS, $query);
            break;
            case 'DELETE':
                if (count($parameters)) throw new Exception("DELETE requests cannot contain a query string");
                curl_setopt($ch, CURLOPT_URL, $url);
                curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'DELETE');  
            break;
            default: throw new Exception('Unknown request type: '.$type);
        }
        
        $result = curl_exec($ch);
        $status = curl_getinfo($ch, CURLINFO_HTTP_CODE);

        if ($debug) echo "<p>Getting status code $status.</p><p>Request result:<br /><pre>".htmlspecialchars($result).'</pre></p>';
        
        if ($status==400) throw new Exception("Bad request. Make sure that the input data is correct and that any numbers are correctly formatted.");
        elseif ($status==401) throw new Exception("Authorization Error. Check that you are using the correct username and the generated API password.");
        elseif ($status==403) throw new Exception("Operation Not Allowed. Check that you are sending from a valid sender number that is registered with the Message Gateway.");
        elseif(!in_array($status, self::$validStatusCodes)) throw new Exception("Invalid status code returned: ".$status);
        
        return $result;
    }
    
    /** 
    * Chooses a Service Host based on current availability.
    * @return The address of a functioning host, or null if none are available.
    */
    public static function getServiceAddress() {
        $records = dns_get_record(self::SERVICE_ADDRESS,DNS_SRV);
        $sortedRecords = array();      
        $maxWeightTable = array();
        foreach ($records as $record) {
            $pri = $record['pri'];
            $maxWeightTable[$pri] += $record['weight'];
            $sortedRecords[$pri][$record['target']] = $maxWeightTable[$pri];
        }
        ksort($sortedRecords);
        $workingHost = null;
        foreach($sortedRecords as $pri => $data) if (!$workingHost) {
            while(!$workingHost && $host = self::pluckRandomWeighted($data)) {
                if (self::isResponding($host)) $workingHost = $host;
            }
        }
        return $workingHost;
    }

    /**
    * Picks a random hostname from the given array, taking into consideration the weights given as array values.
    * The entry will be removed from the array, preventing it from being chosen again.
    * @param array $hosts Associative array where the keys are hostnames and the values are the corresponding weights.
    * @return Chosen key from the array.
    */
    private static function pluckRandomWeighted(array &$hosts) {
        if (!count($hosts)) return null;
        $chosenHost = null;
        $maxWeight = max($hosts);
        
        $r = rand(0,$maxWeight);
        foreach($hosts as $host=>$weight) {
            if (!$chosenHost && $weight>=$r) {
                $chosenHost = $host;
                $chosenHostWeight = $hosts[$host];
                unset($hosts[$host]);
            } else {
                $hosts[$host] -= $weight;
            }
        }
        return $chosenHost;
    }
    
    /**
    * Determines whether a given service host is currently functioning.
    * @param string The hostname to test.
    * @return boolean True if the host is functioning, false otherwise.
    */
    private static function isResponding($hostname) {
        $ch = curl_init();
        $url = 'https://'.$hostname;
        curl_setopt_array($ch, array(CURLOPT_URL=>$url,CURLOPT_NOBODY=>1,CURLOPT_RETURNTRANSFER=>0));
        $res = curl_exec($ch);
        curl_close($ch);
        return $res;
    }
    
    public static function testWeighting() {
        $start = time();
        $num = 100;
        $res = array();
        for($i=0;$i<$num;$i++){
            $res[self::getServiceAddress()]++;
        }
        print_r($res);
        echo "Total time: ".(time()-$start);
    }

}

?>